/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup llSySer
 * @{
 *
 * \file ll_system.h
 *
 *  \ref intro_sec "weAutSys"'/weAut_01' low level system calls and services
 *
 *  This file contains the definitions of \ref intro_sec "weAutSys"'/weAut_01' low level
 *  system calls and services that are normally not used by application /
 *  user software. This is system software and must not be modified for
 *  user or application programs.
 *
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

/** @} */ // Doxygen addtogroup llSySer

#ifndef LL_SYSTEM_H_
#define LL_SYSTEM_H_
#include "we-aut_sys/common.h"

#if USE_BOOTLOADER >= 2
#include "bootloader/boot109.h"
#endif

#include <avr/wdt.h>


/**
 * \addtogroup llSySer
 * @{
 */


/** Initialise system resources after reset or restart
 *
 *  This function essentially calls all other initialisation functions
 *  in appropriate order (while interrupts are disabled). At return
 *  interrupts are enabled.
 *
 *  This function shall never be called directly by user software.
 *  To cause a software restart use  systemAbort(uint8_t cause).
 */
void initSystemRes(void);

void initPorts(void);  //!< low level reset type initialisation of ports

/** Initialisation of status / test LEDs (if any)
 *
 *  \param LEDgnStart start value for the green LED (ON or OFF)
 *  \param LEDrdStart start value for the red LED (ON or OFF)
 */
void initStatusLeds(uint8_t LEDgnStart, uint8_t LEDrdStart);

void initSystTiming(void); //!<  low level clock / timer / tick initialisation
void initProcIO(void); //!< low level reset type initialisation of process I/O


/** The system start
 *
 *  This is the reset entry point (where it all begins). This function is
 *  \ref intro_sec "weAutSys"' system software. It must not be supplied by
 *  user / application software.
 *
 *  User / application software is written by supplying one or more cyclic or
 *  event driven \ref appThr_sec "threads".
 */
int main(void) __attribute__((OS_main));

/** The last reset cause(s)
 *
 *  This is the state of the MCU Status Register read very early at
 *  system restart. <br />
 *  The register will be reset after setting this variable. <br />
 *  This variable must not be modified elsewhere.
 */
extern uint8_t resetCauses;

/** The last main cause as text
 *
 *  This is a text describing the main reset cause according to
 *  \c resetCauses. It will be set at system start. <br />
 *  Not to be modified elsewhere.
 *
 *  The text will be at least 11 characters long including at least one
 *  trailing blank. It is in flash memory (program space). In case of
 *  full \ref USE_BOOTLOADER "bootloader integration" it will be a 32 bit
 *  address value.
 */
#if USE_BOOTLOADER >= 2
extern uint32_t resetCauseText;
#else
extern char const * resetCauseText;
#endif

/** The abort command respectively cause
 *
 *  Values (in Hex) assigned are: <br />
 *  00 : 0perate, run n0rmally <br />
 *  B0 : Block 0peration, stop (no application threads scheduled,
 *        \ref PLCstop)
 *  ED : unExpected watchDog timeout <br />
 *  EC : Exit by abort Command <br />
 *
 *  Any value not 0x00 and not 0xB0 will drive the system through a
 *  complete reset / restart procedure.
 *
 *  If 0x00 or 0xB0 this variable is set by system software if the
 *  corresponding abort conditions are diagnosed. The same can be done by
 *  user / application software if adequate.
 *
 *  \sa setAbortCommand
 */
extern uint8_t abortCommand;

#define PLCstop  0xB0 //!< Abort command:  stop
#define PLCrun   0x00 //!< Abort command:  run (i.e. no abort command at all)
#define WDtiOut  0xED //!< Abort cause:  (unexpected) watchdog timeout
#define abortHMI 0xCE //!< Abort command: abort by human command entry
#define abortBoot 0xCB //!< Abort command: go to bootloader (by human command)

/** Status check: PLC is in Stop */
#define PLCinSTOP (abortCommand == PLCstop)

/** Status check: PLC is in Run */
#define PLCinRUN (!abortCommand)


/** Set the abort cause respectively command
 *
 *  This function will set the the abort \ref abortCommand "command or cause"
 *  to the parameter value as long as the current state is \c PLCstop or
 *  \c PLCrun. <br />
 *  In other words: From state run or stop any state can be set including
 *  an abort / fault state. But an abort / fault state will not be left or
 *  changed by this function.
 *
 *  \param command the new run stop fault abort state / command
 *  \sa abortHMI
 *  \sa abortBoot
 *  \sa PLCrun
 *  \sa PLCstop
 *  \sa WDtiOut
 */
void setAbortCommand(uint8_t command)  __attribute__((always_inline));


/** @} */ // Doxygen addtogroup llSySer


/**
 *  \addtogroup sysSer
 * @{
 *
 * \addtogroup timeKeep
 * @{
 */

/** Prescaled time factor to millisecond
 *
 *  By this factor the processor clock period pre-scaled by 256 will be the
 *  best <= approach to one millisecond.
 *
 *  For a 20 MHz machine the pre-sclaled clock period will be 12.8 s and this
 *  factor 78. That will yield 998.4 s which cal easily be slowed down by
 *  VCO to 1ms.
 *
 *  On a 16 MHz machine that will be 16s, 62 and 992s
 */
#define PRESC_TRIM_FAC ((uint8_t)(F_CPU / 254410))

/* * Prescaled time factor to millisecond (slow)
 *
 *  This is just \ref PRESC_TRIM_FAC + 1.
 *  /
#define PRESC_TRIM_SLOW ((uint8_t)(F_CPU / 254410) + 1)  */


/** The ms tick counter
 *
 *  This variable is written, i.e. incremented, in the tick interrupt
 *  handler \c ISR(TIMER0_COMPA_vect) only. <br />
 *  This variable must never (!) be modified elsewhere.
 *
 *  Being just one byte reading it is atomic.
 *  \sa syst_threads.h
 */
extern uint8_t msIntTick;

/** The 12,8 s counter summand
 *
 *  Incremented in the 1ms interrupt by \ref PRESC_TRIM_FAC (or
 *  \ref PRESC_TRIM_FAC + 1 if VCO \ref slowVCOdown "slow").
 *
 *  This variable is not to be used directly by user / application software.
 *  Use it idirectly by \ref cnt12u8_8().
 */
extern uint8_t cn12u8;


/** Trimming of timing oscillator
 *
 *  The higher the value the faster the oscillator for the system 1 ms ticks
 *  will be running (in the mean). The relation is non-linear.
 *
 *  Background: This value is used to slow down respectively lengthen every
 *  1st .. 255th millisecond tick period a tiny bit. All others will be a
 *  wee bit too fast.
 *
 *  Range: 1 .. 255 (0 would act as 256)
 *
 *  The best value for a 16 MHz machine is 2.
 *  The best setting for \ref intro_secH "weAut_01" board
 *  with a good 20 MHz Quartz for ATmega1284's processor clock is 7.
 *
 *  \showinitializer
 */
#if F_CPU == 16000000
#define VCO_DEFAULT 2
#else
#define VCO_DEFAULT 7
#endif

/** Trimming of timing oscillator
 *
 *  This function will set the "value controlled oscillator" for the
 *  system's 1 ms tick to normal (usually best) speed.
 *
 *  \note If  \ref intro_sec "weAutSys"'s NTP client is active this
 *  function is for its exclusive use only.
 *  \sa VCO_DEFAULT
 */
void setVCOnormal(void) __attribute__ ((always_inline));

/** Trimming of timing oscillator
 *
 *  This function returns the current speed of the
 *  "value controlled oscillator" for the
 *  system's 1 ms tick. \ref VCO_DEFAULT is usually normal / best setting;
 *  higher values are faster.
 */
uint8_t getVCOsetting(void) __attribute__ ((pure));


/** Trimming of timing oscillator
 *
 *  This function will speed up the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below -200 ms.
 *
 *  See the note at \ref setVCOnormal.
 *  \sa VCO_DEFAULT
 */
void speedVCOup(void) __attribute__ ((always_inline));

/** Trimming of timing oscillator
 *
 *  This function will slow down the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below +100 ms.
 *
 *  See the note at \ref setVCOnormal.
 *  \sa VCO_DEFAULT
 *  \sa setVCOnormal
 */
void slowVCOdown(void) __attribute__ ((always_inline));

/** Get the 12.8 (16) s tick value (8 bit)
 *
 *  This function returns the actual 12.8 s respectively 16 s tick count.
 *  This value will wrap about every 3.27 ms on a 20 MHz machine and every
 *  4 ms at 16 MHz. One tick (increment) is 256 processor clocks.
 *
 *  It might be used for quite exact measures of software execution times the
 *  normal \ref msIntTick "milliseconds" resolution is of little use for.
 *
 *  Hint: 12.8 s is correct for a 20 MHz machine. For slower machines the
 *  counted period is proportionally longer. For 16 MHz it's 16 s.
 *  \sa PRESC_TRIM_FAC
 */
uint8_t cnt12u8_8(void) __attribute__ ((pure));


/** @} */ // Doxygen addtogroup timeKeep

/**
* \addtogroup sysThr
* @{
*/


#ifndef NO_SPMIN_SAMPLE
/** Do not sample the stack pointer and update its minimal value
 *
 *  To hinder the \ref SAMPLE_MIN_SP "sampling" the
 *  \ref minStckP "minimal stackpointer" define the macro \ref NO_SPMIN_SAMPLE
 *  with a value of 1.
 *  \showinitializer
 */
#define NO_SPMIN_SAMPLE 0
#endif
/** Sample the stack pointer and update its minimal value
 *
 *  The \ref minStckP "minimal stackpointer" will be updated at the point
 *  of calling this, if the macro \ref NO_SPMIN_SAMPLE is undefined or 0.
 *  \param nonc number of nested function calls within embedding function <br />
 *              The value should be in the range 0..3. <br />
 *              If the call hierarchy goes deeper or the functions called
 *              use local variables those
 *              functions called should \ref  SAMPLE_MIN_SP "sample" self.
 */
#if NO_SPMIN_SAMPLE
# define SAMPLE_MIN_SP(nonc) do { } while(0)
#else
# ifdef __AVR_3_BYTE_PC_
#  define SAMPLE_MIN_SP(nonc) do { \
   uint16_t sampSP = SP - (3 * (nonc)); \
   if (sampSP < minStckP) minStckP = sampSP; \
   } while(0)
# else
#  define SAMPLE_MIN_SP(nonc) do { \
   uint16_t sampSP = SP - (2 * (nonc)); \
   if (sampSP < minStckP) minStckP = sampSP; \
   } while(0)
# endif
#endif

/** Minimal stack pointer value sampled
 *
 *  System and application software might sample the stackpointer at due
 *  points and set this variable to it if larger. A value of 0xFFFF means
 *  no (new) samples taken as yet.
 *  The recipe to \ref SAMPLE_MIN_SP "sample" the stackpointer \c SP
 *  \ref SAMPLE_MIN_SP "and update" its minimal value is \code
    uint16_t value = SP;
    if (value < minStckP) minStckP = value; // sample stackpointer \endcode
 */
extern uint16_t minStckP;

/** Go to a label (as value) and stay within the 64Kword page
 *
 *  This procedure is a helper to implement Adam Dunkels' Protothreads with
 *  labels as values on large ATmegas with more than 128K flash.
 *
 *  Hint / Warning: This does only work (respectively has to be used) as long
 *  as a) there are no 24 bit void pointers and b) the trampolin (+ linker
 *  relaxation) hack is not in effect. As of January 2014 both a) and b) are
 *  true for arv-gcc &mdash; hence do not use this (see \ref TRAMPOLIN_USE).
 *
 *  \param labelAsValue the target (jump) address' lower 16 bit
 *  \return byte 3 of return address or 0
 */
void goto_P(void * labelAsValue);

/** Get the upper (bit 16 ...) part of the current address
 *
 *  This function is a helper to implement Adam Dunkels' Protothreads with
 *  labels as values on large ATmegas with more than 128K flash.
 *
 *  \return byte 3 of return address or 0
 *  \sa goto_P
 *  \sa addrHere
 */
uint8_t addr3Here(void);

/** Get the current address
 *
 *  This function returns the address from where it was called &mdash; or to
 *  be exact, the flash address behind the instruction calling this function.
 *
 *  \return calling address +2; the number of relevant bits depends
 *          on the flash size
 */
uint32_t addrHere(void);

/** @} */ // Doxygen addtogroup sysThr
/** @} */ // Doxygen addtogroup sysSer
#endif //  LL_SYSTEM_H_
/** @} */ // Doxygen addtogroup weAutSys

