// Copyright (c) 2011 Albrecht Weinert, Bochum *
// $Id: uart0.h 2 2017-01-25 16:49:03Z albrecht $

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 * \addtogroup serCom
 * @{
 */

/**
 *  \file uart0.h
 *
 *  Serial communication (basics)
 *
 *  This file contains the definitions for the basics of
 *  \ref intro_sec "weAutSys"' serial communication. Please find detailed
 *  description of the variables, functions etc. in the modules
 *  \ref serCom "Serial communication" and \ref serDrv "Serial com. drivers".
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $
/** @} */ // Doxygen addtogroup sysSer
/** @} */ // Doxygen addtogroup serCom

#ifndef __UART0_H__
#define __UART0_H__

#include <avr/pgmspace.h>
#include "arch/config.h"

/**
 * \addtogroup llSySer
 * @{
 * \addtogroup serDrv
 * @{
 */

// -- output ---

/** The serial output buffer */
extern uint8_t uartOutBuf[];

/** The serial output buffer write/put index */
extern volatile uint8_t uartOutBufWi;

/** The serial output buffer read/get index */
extern volatile uint8_t uartOutBufRi;


/** Clear the internal buffer for serial output
 *
 *  This function returns the number of characters forgotten.
 *  If 0 is returned the buffer was empty anyway.
 *
 *  This function is to be used to recover from buffer overrun or similar
 *  errors or to "forget the past" out of any other reasons.
 */
uint8_t uartClearOutBuffer(void);

/** Driver helper
 *
 *  This is a low level helper macro for driver programming. <br />
 *  Warning: It is to be used by system software only and
 *  not outside atomic locks.
 */
#define FROM_BUFFER2UART() { \
   UDR0 = uartOutBuf[uartOutBufRi]; \
   uartOutBufRi = (uartOutBufRi + 1) & UART_OUT_BUF_CAP; \
}

/** UART (0) can get send data
 *
 *  This evaluates to non 0 (true), if the transmit data register
 *  is empty.
 */
#define UART_CAN_SEND (!(isRTSflowC && (PINB & 2)) && (UCSR0A & (1<<UDRE0)))


/** The serial output buffer is not empty
 *
 *  This evaluates to true, if the serial output buffer is NOT empty.
 */
#define SER_SENDBUF_NOT_EMPTY (uartOutBufWi != uartOutBufRi)

/** The serial output buffer is empty
 *
 *  This evaluates to true, if the serial output buffer is empty.
 */
#define SER_SENDBUF_EMPTY (uartOutBufWi == uartOutBufRi)

//---- out / in ------

/** The serial input buffer is empty
 *
 *  This evaluates to true, if the serial input buffer is empty.
 */
#define SER_RECVBUF_EMPTY (uartInBufWi == uartInBufRi)

/** The serial input buffer is not empty
 *
 *  This evaluates to true, if the serial input buffer is NOT empty.
 */
#define SER_RECVBUF_NOT_EMPTY (uartInBufWi != uartInBufRi)


/** @} */ // Doxygen addtogroup serDrv
/** @} */ // Doxygen addtogroup  llSySer

/**
 * \addtogroup sysSer
 * @{
 * \addtogroup serCom
 * @{
 */

/** The capacity of the serial output buffer
 *
 *  The value must be 2**n-1 and fit into an unsigned byte. <br />
 *  Possible values are 15, 31, 63, 127 and 255. <br />
 *  Recommended (default) is 255. This is a good compromise concerning RAM
 *  saving and avoiding thread waits / yields due to buffer fill. Use lower
 *  values only if the RAM is urgently needed.
 *
 *  The capacity's upper limit (255) is fixed by the type (uint8_t) chosen for
 *  lengths and indices. This makes most respective operations more than twice
 *  as fast as would be with 16 bit types (uint16_t). If larger UART buffers
 *  should become necessary that could be changed on customer request. As an
 *  alternative using higher baud rates should be considered. At (default)
 *  38400 it takes 67 ms to empty (send) a 256 byte buffer. That is reduced to
 *  22ms with 115K2. Sending more than 100 characters over the UART (every
 *  time) in the \ref app10msThread "10 ms thread" (or 23 at 38400) would fail
 *  anyway in the long run, no matter the buffer size. Consider
 *  \ref ethCom_sec "Ethernet communication" (probably utilising the
 *  \ref telnetServer "Telnet") as alternative.
 *
 *  \showinitializer
 *  \sa UART_IN_BUF_CAP
 */
#define UART_OUT_BUF_CAP 255


/** The buffer space available for serial output
 *
 *  This function returns the number of bytes that can be output UART
 *  output by \ref serPutChar() e.g. without loosing data.
 *
 *  The returned value is in the range 0 ..\ref UART_OUT_BUF_CAP - 1. <br />
 *  If 0 is returned nothing should be output.
 */
uint8_t uartOutSpace(void);


/** Put some characters to serial output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller than \c n or than the length of the string \c src. It may 
 *  even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref uartOutSpace().
 *
 *  The output stops after n characters or at the first 0 in src. By that
 *  this is a string output function. To send a (binary) 0 use
 *  \ref serPutChar(char c, ...) or \ref uartPutBytes().
 *
 *  \param src the characters to be output (not NULL !)
 *  \param n   the maximum number of characters to be output
 *  \return the number of characters output
 */
uint8_t uartPutChars(char * src, uint8_t n);

/** Put some bytes (characters) to serial output
 *
 *  This function never blocks. It returns the number of bytes
 *  transferred. If not enough space is available the returned value will
 *  be smaller than \c n. It may even be 0. In the latter case nothing was
 *  output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref uartOutSpace().
 *
 *  \param src the characters to be output
 *  \param n   the maximum number of characters to be output
 *  \return the number of characters output
 */
uint8_t uartPutBytes(uint8_t * src, uint8_t n);

/** Put a RAM string (some characters) to serial output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller the length of the string \c src. It may even be 0.
 *  In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref uartOutSpace().
 *
 *  The output stops after n characters or at the first 0 in src. By that
 *  this is a string output function. To send a (binary) 0 use
 *  \ref serPutChar(char c, ...).
 *
 *  Hint:  The parameter stream is not used in the present (June 2012)
 *    implementation, but should be set correctly in case of future changes.
 *
 *  \param src the characters to be output
 *  \param stream the serial output as stream
 *  \return the number of characters output
 */
uint8_t uartPutSt(char * src, const FILE * stream);


/** Put a string (some characters) from flash memory to serial output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller the length of the string \c src. It may even be 0. In the
 *  latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref uartOutSpace().
 *
 *  The output stops after n characters or at the first 0 in src. By that
 *  this is a string output function. To send a (binary) 0 use
 *  \ref serPutChar(char c, ...).
 *
 *  Hint:  The parameter stream is not used in the present (June 2012)
 *    implementation, but should be set correctly in case of future changes.
 *
 *  \param src the characters to be output
 *  \param stream the serial output as stream
 *  \return the number of characters output
 */
int uartPutSt_P(prog_char * src, const FILE * stream);


/** Put some characters from program space to serial output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller than \c n or the length of the string \c src. It may even
 *  be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by
 *  \ref uartOutSpace().
 *
 *  The output stops after n characters or at the first 0 in src.
 *
 *  \param src the characters to be output in flash memory (not NULL !)
 *  \param n   the maximum number of characters to be output
 *  \return the number of characters output
 */
int uartPutCharsP(char const * src, uint8_t n);


//--------------------  simple UART in functions   ----------------------

/** Get one byte from serial input
 *
 *  This function never blocks. It returns a byte value read or EOF. <br />
 *  Hence this operation should be (indirectly) guarded by uartInBufferd().
 *
 *  \param stream (passed by stdio input) totally ignored by this function
 *  \return a character 0..255 or EOF
 */
int uartGetChar(FILE *stream);


//--------------------  end simple UART functions  -----------------------

/** @} */ // Doxygen addtogroup serCom
/** @} */ // Doxygen addtogroup sysSer

// -- input --

/**
 * \addtogroup llSySer
 * @{
 * \addtogroup serDrv
 * @{
 */

/** The serial input buffer  */
extern uint8_t uartInBuf[];

/** The serial input buffer write/put index */
extern uint8_t uartInBufWi;

/** The serial input buffer read/get index  */
extern uint8_t uartInBufRi;

/** @} */ // Doxygen addtogroup serDrv
/** @} */ // Doxygen addtogroup llserCom


/**
 * \addtogroup sysSer
 * @{
 * \addtogroup serCom
 * @{
 */

/** The maximum capacity of the serial input buffer
 *
 *  The value must be 2**n-1 and fit into an unsigned byte. <br />
 *  Possible values are 15, 31, 63, 127 and 255. <br />
 *  Recommended (default) is 127 that is fit for terminal command lines.
 *
 *  It may be set to to 255 if RAM is available. For still larger buffer sizes
 *  compare the remarks about the  \ref UART_OUT_BUF_CAP "output buffer"
 *  capacity. If input size is critical \ref isCTSflowC "input flow control"
 *  is the recommended means to put the buffer burden to the sender.
 *
 *  \showinitializer
 *  \sa UART_OUT_BUF_CAP
 */
#define UART_IN_BUF_CAP 127

/** UART space limit for flow control
 * \showinitializer
 * \sa B1D6MODE_CTS_FLOWC
 */
#define UART_IN_SPACE_LIM 11


/** The serial input buffer line feeds code counter */
extern uint8_t uartInBufRetCnt;

/** The serial input's accumulated (or) errors
 *
 *  If 0 no errors had occurred since last check.
 *
 *  Bit 4 set: at least one framing error had occurred. <br />
 *  Bit 3 set: at least one overrun error had occured. <br />
 *  Bit 2 set: at least one parity error had occurred. <br />
 *  Bit 6 set: at least one byte could not be put to
 *     \ref uartInBufferd "buffer". <br />
 *  Bit 5 set: at least one \ref UART0SPIYKY "spike" was ignored. <br />
 */
extern uint8_t uartInAccErrors;

/** The number of characters buffered from serial input
 *
 *  This function returns the number of bytes that can be fetched from
 *  serial input by \ref uartGetChar(FILE*) without getting EOF.
 *
 *  The returned value is in the range 0 .. \ref UART_IN_BUF_CAP - 1. If 0 is
 *  returned nothing should be read. If \ref UART_IN_BUF_CAP is returned the
 *  internal buffer was overrun and the input data are incomplete
 *  respectively corrupt.
 */
uint8_t uartInBufferd(void);

/** The number of characters buffered from serial input
 *
 *  Same as \ref uartInBufferd (other respectively common signature).
 */
uint16_t serInBufferd(FILE * streams);

/** The number of line feeds buffered from serial input
 *
 *  This function gives the number of line feeds, i.e. \c LF or \c CR chars
 *  that were read from serial input and not yet fetched from buffer by
 *  uartGetChar(FILE*) or uartGetLine(char[], int).
 *
 *  The returned value is in the range 0 .. UART_IN_BUF_CAP . If 0 is
 *  returned no (stdio) function expecting a complete line input should
 *  be called.
 */
uint8_t uartInRetBufferd(void);

/** The accumulated serial input (UART0) errors
 *
 *  This function returns all UART 0 input errors accumulated (ORed) in the
 *  past: <br />
 *  Bit 6: buffer overrun. More than UART_IN_BUF_CAP bytes were received
 *         but not fetched in due time. <br />
 *  Bit 4: framing error. This usually happens if the two UARTs connected do
 *         not have the same setting, one of their clocks are out of
 *         tolerance, hardware noise or connector troubles. <br />
 *  Bit 3: UART buffer overrun. This should not happen, the effect of
 *         slow input processing is Bit 6. If this happens, some interrupt
 *         handlers or (atomic) critical sections block interrupts too
 *         long. <br />
 *  Bit 2: Parity error. This has often the same causes as bit 4. Can only
 *        happen with parity enabled. <br />
 *
 *  If this function returns 0 all was well since last call or last
 *  buffer clearance. A call to this function resets the errors.
 */
uint8_t uartInErrors(void);


/** Clear the internal buffer for serial input
 *
 *  This function returns the number of characters forgotten.
 *  If 0 is returned the buffer was empty anyway.
 *
 *  This function is to be used to recover from buffer overrun or similar
 *  errors or to "forget the past" out of any other reasons.
 */
uint8_t uartClearInBuffer(void);

/** Read a line from UART
 *
 *  This function will copy all characters to the array (\c line) supplied
 *  ending with line feed. This line feed may be one or two characters LF
 *  or CR; it will be put as one ending 0 as last character into line.
 *
 *  Returned is the number of characters copied to the array \c line.
 *  This might be 0; it will never be more than \c max,
 *  \ref uartInRetBufferd() and \ref UART_IN_BUF_CAP.
 *
 *  A full line was got if the last character ([returned value -1]) is
 *  (char) 0 which will, as said, replace the terminating line feed form the
 *  (buffered) input. So, space permitting, the modified \c line[] will be a
 *  null-terminated (C-) string.
 *
 *  To guarantee a full line, a call to this function should be guarded
 *  by  \ref uartInRetBufferd()  and max better be >= UART_IN_BUF_CAP.
 *  \param line the character array to put the received line to
 *  \param max  the maximum number of characters transferred
 *  \return the number of characters transferred
 *  \sa uartGetCmdLine
 */
uint8_t uartGetLine(char * line, uint8_t max);


/* Read a command line from UART
 *
 * This function does essentially the same as
 * \ref uartGetLine "uartGetLine(line,  max)". Additionally it will filter
 * out all leading white spaces and (non ASCII) characters above 'z'.
 *
 * The rationale for this filtering is <br />
 * a) the rule that CLI commands should be restricted to ASCII anyway and
 * b) some PC UART drivers (Linux and Windows) tend to send spikes when
 * opening / closing a com port or when entering or leaving a power save
 * mode.
 *
 * Those spikes will eventually be seen by the receiver as a start bit + 0..4
 * one bits.
 *  \param line the character array to put the received line to
 *  \param max  the maximum number of characters transferred
 *  \return the number of characters transferred
 */
uint8_t uartGetCmdLine(char * line, uint8_t max);

/** Sets UART flow control by (one) character
 *
 *  i c: input flow control by CTS <br />
 *  o r: output flow control by RTS <br />
 *  b  : both CTS and RTS
 *  n, all else: no flow control <br />
 *
 *  \param cC the controlling character (may also be in upper case)
 *  \sa setB1D6mode
 */
void setUARTflowcontrolByChar(char cC);


/** @} */ // Doxygen addtogroup serCom
/** @} */ // Doxygen addtogroup sysSer

//--------------   both / initialise      ----------------------------


/**
 * \addtogroup llSySer
 * @{
 * \addtogroup  sysInit
 * @{
 */

/* * Calculate the prescaler setting for a desired baudrate
 *
 *  The calculated and returned \c prescaler divisor setting must be in the
 *  range 0 .. 4096. Otherwise the desired \c baudRate can't even be
 *  approximated with the given processor frequency (e.g. F_CPU=20000000).
 *
 *  The effective baudRate can be calculated by
 *  \ref uartBaud(uint16_t, uint8_t) to check the tolerance/error.
 *
 *  \note This function probably involves 32 bit arithmetic making it a bit
 *  expensive with respect to processor usage. It is intended for use in the
 *  reset / initialisation phase. Instead of calling it regularly with a
 *  limited set of parameter values, it is recommendable to cache those
 *  repeatedly needed results.
 *
 *  \param x2 true: double speed (half receive samples)
 *  \param baudRate the rate desired
 *  \return baudDivide, the prescaler setting, must be 0..4096
 * /
uint16_t uartPrescaler(uint32_t baudRate,  uint8_t x2); */

/** Set the UART0's baudrate divisor
 *
 *  This function sets the baud rate prescaler register and hence the 
 *  baud rate.
 *
 *  It is normally called indirectly while initialising.
 *  Calling while running may spoil ongoing transmissions.
 *
 *  \param baudDivide 0..4096
 *  \sa uartInit
 */
void uartSetBaudDivide(uint16_t baudDivide) __attribute__((always_inline));


/** Initialise the serial input (UART0)
 *
 *  This function is normally used after system reset. <br />
 *  To call this function while the system is running is not recommended
 *  or should at least be considered and tested very carefully.
 *
 *  UART 0 is initialised in asynchronous mode for receiving and
 *  transmitting (full duplex).
 *
 *  \param baudDivide 0..4096 prescaler
 *  \param x2 true: double speed (half receive samples)
 *  \param len 5,6,7, 8 [default] (9 [not yet supported])
 *  \param parity 0=false: none; odd , even value: odd, even
 *  \param stopBits 2: 2 stop bits, 1 [default]: 1 stop bit
 *  \sa uartSetBaudDivide
 */
void uartInit(uint16_t baudDivide,  uint8_t x2,
                             uint8_t len, uint8_t parity, uint8_t stopBits);

/** Get the usage of the ports PD6 and PB1
 *
 *  These Controller ports can be used in a variety of ways independent of
 *  the serial communication link. In the weAut_01 module PD5 can be jumpered
 *  to the CTS output via MAX202 level shifter / inverter and PB1 can be
 *  jumpered to the RTS input.
 *
 *  If so connected they may be used for input (RTS-) CTS- flow control if
 *  this mode is set so.
 *  \sa setB1D6mode
 */
uint8_t getB1D6mode(void) __attribute__((always_inline));

/** Usage mode of ports PD6 and PB1: none
 *
 *  Application software may set and use these ports freely.
 *  \showinitializer
 */
#define B1D6MODE_NONE 0

/** Usage mask for PB1
 *
 *  The upper four bits of \ref getB1D6mode "b1d6Mode" concern Port B1.
 *  If non of those bits is set the application software may set and use
 *  this port freely &mdash; i.e. according to hardware / jumpers.
 *  \showinitializer
 */
#define B1MODE_MASK 0xF0

/** Usage mask for PD6
 *
 *  The lower four bits of \ref getB1D6mode "b1d6Mode" concern Port D6.
 *  If non of those bits is set the application software may set and use
 *  this port freely &mdash; i.e. according to hardware / jumpers.
 *  \showinitializer
 */
#define D6MODE_MASK 0x0F

/** Usage mode of port PD6 : CTS input flow control
 *
 *  The port PD6 must be connected to the CTS output via a level shifter /
 *  inverter (MAX202 e.g.).
 *
 *  In this mode the UART software will signal their ability to receive more
 *  bytes according to \ref UART_IN_SPACE_LIM (off) resp.
 *  (3 * UART_IN_SPACE_LIM) (om).
 *  \showinitializer
 */
#define D6MODE_CTS_FLOWC 0x01

/** Usage mode of port PD6 : CTS input flow control
 *
 *  \sa D6MODE_CTS_FLOWC
 *  \sa getB1D6mode()
 */
#if UART0MAY_CTS
#define isCTSflowC ((getB1D6mode() &  D6MODE_MASK) == D6MODE_CTS_FLOWC)
#else
#define isCTSflowC 0
#endif

/** Usage mode of port PB1 : RTS output flow control
 *
 *  The port PB1  must be connected to the RTS input via a level shifter /
 *  inverter (MAX202 e.g.).
 *
 *  In this mode the UART software will send only if the other station sets
 *  RTS to logic 0 (+12V).
 *
 *  \note
 *  This is a non-standard but nevertheless (only) reasonable symmetric
 *  usage of RTS / CTS. CTS is set here independent of RTS and not as
 *  response to it. CTS, is set by \ref D6MODE_CTS_FLOWC to be used just
 *  indicates permission this device to send data to it. RTS indicates
 *  permission from the other station to this device to send. Hence it would
 *  better be called RTR (Ready to Receive) in the other stations point of
 *  view.
 *
 *  Two (weAut_01 / \ref intro_sec "weAutSys") devices with both set and jumpered to
 *  \ref B1D6MODE_RTSCTS_FLOWC may very well be P2P connected by serial link.
 *
 *  \showinitializer
 */
#define B1MODE_RTS_FLOWC 0x10

/** Usage mode of port PB1: RTS output flow control
 *
 * \sa B1MODE_RTS_FLOWC
 * \sa getB1D6mode()
 */
#if UART0MAY_RTS
#define isRTSflowC ((getB1D6mode() &  B1MODE_MASK) == B1MODE_RTS_FLOWC)
#else
#define isRTSflowC 0
#endif

/** Usage mode of ports PD6 and PB1: RTS CTS input flow control
 *
 *  The port PD6 must be connected to the CTS output and PB1 to the RTS
 *  input &mdash; both via level shifters / inverters (MAX202 e.g.).
 *
 *  In this mode the UART software will use those ports for input and
 *  output flow control and all other settings and uses by application
 *  software are forbidden interference.
 *
 *  This is, of course, just \ref D6MODE_CTS_FLOWC and
 *  \ref B1MODE_RTS_FLOWC combined
 *  \showinitializer
 */
#define B1D6MODE_RTSCTS_FLOWC 0x11


/** Set the usage of the ports PD6 and PB1
 *
 *  \sa getB1D6mode
 *  \sa B1D6MODE_RSTCTS_FLOWC
 *  \sa B1D6MODE_NONE
 *  \param mode the new mode
 *  \sa setUARTflowcontrolByChar
 */
void setB1D6mode(uint8_t mode);

/** @} */ // Doxygen addtogroup sysInit
/** @} */ // Doxygen addtogroup llSySer

#endif /* __UART0_H__ */

/** @} */ // Doxygen addtogroup weAutSys
