/**
 *  \file ll_system.c
 *
 *  weAutSys'/weAut_01' low level system calls and services
 *
 *  This file contains the implementation of weAutSys'/weAut_01' low level
 *  system calls and services. that are normally not used by application /
 *  user software. This is system software and must not be modified for
 *  user or application programs.
 *
 * This file is part of weAutSys &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include "arch/config.h"
#include "we-aut_sys/common.h"

#if USE_BOOTLOADER >= 2
#include "bootloader/boot109.h"
#endif


/** This function does nothing
 *
 *  It is just a hack to check sizes (\ref EXTRA_THR_ST_SZ,
 *  \ref SIZE_OF_BIGGEST_APPSTATE etc.) at compile time, i.e. to do something
 *  the preprocessor can't. If all is well the GCC compiler / code generator
 *  will optimize the function body away. Hence this does no harm.
 */
void tuNix(void){
   // if this fails make EXTRA_THR_ST_SZ larger
   ((void)sizeof (char[EXTRA_THR_ST_SZ - sizeof(struct cliThr_data_t)]));

   // if this fails make SIZE_OF_BIGGEST_APPSTATE larger
   ((void)sizeof (char[SIZE_OF_BIGGEST_APPSTATE - sizeof(struct thr_data_t)]));
} // dummy doo

#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/system.h"
#include "maketime_utc.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/spi.h"
#include "we-aut_sys/uart0.h"     // for debug only
#include "we-aut_sys/streams.h"
#include "we-aut_sys/log_streams.h"
#include "we-aut_sys/proc_io.h"
#include "we-aut_sys/persist.h"
#include "we-aut_sys/enc28j60.h"
#include "we-aut_sys/smc.h"

#include "uip/uip.h"
#include "uip/uip_arp.h"

#include "pt/pt.h" // for sake of Eclipse only

struct pt ptAppInitThread; //!< raw initialisation protothread's structure

uint8_t resetCauses;   //!< the last reset cause(s)

#if USE_BOOTLOADER >= 2
uint32_t resetCauseText;
#else
char const * resetCauseText;
//                                0123456789.123456789v
INFLASH(char const sysResetCause1[]) = "power on   "; //!< reset cause: power on
INFLASH(char const sysResetCause2[]) = "external reset "; //!< reset cause: external
INFLASH(char const sysResetCause4[]) = "power (brown) out "; //!< reset cause: brown out
INFLASH(char const sysResetCause8[]) = "watchdog reset    "; //!< reset cause: watchdog
INFLASH(char const sysResetCauseG[]) = "jtag reset "; //!< reset cause: JTAG
/** reset cause: exit from from active bootlaoader or unknown */
INFLASH(char const sysResetCause0[]) = "ex bootloader (or?) ";
#endif


/*  The abort command respectively cause
 *
 *  Values (in Hex) assigned are:<br />
 *  00 : 0perate, run n0rmally <br />
 *  B0 : Block 0peration, stop (no application threads scheduled,
 *        \ref PLCstop) <br />
 *  ED : unExpected watchDog timeout <br />
 *  EC : Exit by abort Command <br />
 */
uint8_t abortCommand;

/* Set the abort cause respectively command
 *
 *  This function will set the the \c abortCommand to a value other than
 *  \c PLCstop and \c PLCrun if not yet set so.
 */
void setAbortCommand(uint8_t command){
   if (!abortCommand || abortCommand == PLCstop) {
      abortCommand = command;
   }
} // setAbortCommand(uint8_t)

INFLASH(char txtAbd[]) = "abort/reset "; //!< report text: abort/reset
//                        0123456789.12
INFLASH(char txtWDt[]) = "watchdog TO "; //!< report text: watchdog time out

struct mThr_data_t  app1msThread;
struct mThr_data_t  app10msThread;
struct mThr_data_t  app100msThread;
struct mThr_data_t  app1sThread;
struct thr_data_t   appSerInpThread;
struct mThr_data_t  appBgTskThread;


/*  The system start
 *
 *  This is the reset entry point (where it all begins).
 */
int main(void){
   cli();
   resetCauses = MCUSR; MCUSR = 0; // save reset cause
   abortCommand = 0x00;  // run
   wdt_disable();    cli();        // disable watchdog for initialisation
#if HAVE_NIC
   networkNotReady = ENC_NRD | ENC_HWP | STC_NST;
#endif


#if USE_BOOTLOADER >= 2
   resetCauseText = resetCauseText_P(resetCauses);
#else
   if (! resetCauses) {
      resetCauseText = sysResetCause0;
   } else if (resetCauses & 8) {
      resetCauseText = sysResetCause8;
   } else if (resetCauses & 16) {
      resetCauseText =  sysResetCauseG;
   } else if (resetCauses & 1) {
      resetCauseText = sysResetCause1;
   } else if (resetCauses & 2) {
      resetCauseText = sysResetCause2;
   } else if (resetCauses & 4) {
      resetCauseText = sysResetCause4;
   } // resetCauses
#endif

   initStdStreams(&bufLogStreams); // stdio streams
   bufLogStreamsInit(); // init log stream
   bufLogSetConsumer(0); // log stream no consumer

#if HAVE_NIC
   spi2EtherChipBaud = SPI5MHz; // SPI(2) clock frequency for Ethernet chip
#endif
#if HAVE_SMC
   spi2MemCardBaud =  SPI5MHz; // SPI(2) clock frequency for memory card
   initSMCthreadState(0x80); // initialise SMC handling thread
#endif

   PT_INIT(&ptAppInitThread); // initialise user init thread
   ptfnct_t appInitThrRet = appInitThreadF(&ptAppInitThread);  // phase 1

   initSystemRes();
   if (appInitThrRet < PT_EXITED ) {
      appInitThrRet = appInitThreadF(&ptAppInitThread);  // phase 2
   }
   persistInit();

   if (appInitThrRet < PT_EXITED ) {
      appInitThrRet = appInitThreadF(&ptAppInitThread);  // phase 3
   }

#if HAVE_NIC
   networkInit();
   lanComStdInit();
#endif

   if (appInitThrRet < PT_EXITED ) {
      appInitThrRet = appInitThreadF(&ptAppInitThread);  // phase 4
   }

   // phase 5 // empty

   // phase 5 // reschedule until finished

   while (appInitThrRet < PT_EXITED ) {
      appInitThrRet = appInitThreadF(&ptAppInitThread);  // phase 5
   }

   wdt_enable(WDTO_250MS);  // watchdog 1/4s. can be extended to WDTO_2S
   WDTCSR |= (1 << WDIE);  // + interrupt


   while(1) { // live loop
      sys1msThread();   // the ms system thread first of all
   //    sys10msThread();   // the 10 ms system thread next // embed in 1ms
      sys100msThread(); // the 100 ms system thread next
      sys1sThread();   // the 1 s system thread next
      // all system threads were scheduled

      if (PLCinSTOP) {
         scheduleYldAppThread(appSerInpThread); // ser. input may get command
         scheduleRegdAppThread(appBgTskThread);
         continue;  // omit other user / application threads
      }
      if (abortCommand) break; // exit live loop on abort command / cause

      // todo rework scheduling
      // status: just schedule all registered app threads
      scheduleRegdAppThread(app1msThread);
      scheduleRegdAppThread(app10msThread);
      scheduleRegdAppThread(app100msThread);

      scheduleYldAppThread(app1sThread);
      scheduleYldAppThread(appSerInpThread);
      scheduleRegdAppThread(appBgTskThread);   // todo do not schedule uncond.
   } // live loop

   // to here on abort only (abortCommand != 0 && != PLCstop)

   //  Abort the system
#if USE_BOOTLOADER
   if (abortCommand == abortBoot) {
      cli();
      asm volatile ("clr __zero_reg__");
      asm volatile ("ldi r24,%0":: "M" (RAMEND & 0xFF));
      asm volatile ("ldi r25,%0":: "M" (RAMEND >> 8));
      asm volatile ("out __SP_H__,r25" ::);
      asm volatile ("out __SP_L__,r24" ::);

      ((void (*)())BOOTL_BEG)(); // enter the bootloader

   } // enter the bootloader
#endif

#if defined(weAut_01) || defined(weAut_00)
   selectDOdrivDIleds(); // select DO & DI-LEDs
   SPDR = 0;  // send one 0 to all selected SPI 1 devices;
#else
#if defined(DI_LEDs0_PORT)
   toPort(DI_LEDs0_PORT,0);
#endif // DI_LEDs0_PORT
#if defined(DO_0_PORT)
   toPort(DO_0_PORT,0);
#endif //  DO_0_PORT
#endif

   uartClearOutBuffer();
#if defined(weAut_01) || defined(weAut_00)
   deSelectSPI1_impl();  // unconditionally de-select (ready after ua...
#endif

   bufLogTxt_P(systBye);
   bufLogTxt_P(txtAbd);

   char s[] = "<? \n\n\n";
   if (abortCommand) twoHexs(s, abortCommand); // 0 = unknown keeps <?
   bufLogTxt(s, 6);



   enableDOdriver(OFF);

   // spin until UART empty
   while (uartOutSpace() != UART_OUT_BUF_CAP) { } // should abort

   //  an endless loop with interrupts disabled (will drive into WD reset)
   do { cli(); } while (1); // do nothing should reset
   return 1; // come never to here
} //  main


void initSystemRes(){
   cli();
   initPorts();  // platform depending port init
   initStatusLeds(ON, OFF);
//   initTestPins(ON, ON);
   initSystTiming();

#if USE_BOOTLOADER < 2 || ! (defined(weAut_00) || defined(weAut_01))
   // todo ?? do not unconditionally init / use UART
   uartInit(uartPrescaler(38400, OFF), OFF, 8, 0, 2);
  //  bufLogSetConsumer(1); // log stream no consumer
#endif

   initProcIO();

   #if HAVE_NIC
   encReset(); //  ENC reset  clock nr  stack not init'd yet
#endif

   initThreads();
   sei(); // enable interrupts
} //  initSystemRes()


/* low level reset type initialisation of ports */
void initPorts(){

#if defined(weAut_00) || defined(weAut_01)

# if USE_BOOTLOADER < 2
   SPCR =  0; // turn off SPI for (re-)init (essential)
   DDRA  = 0;    // Port A all inputs DI or AI
   PORTA = 0;   // no pullup (weAut_00 has filters)

   // weAut  Port B
   // B0 in   enter key                    in  pullup               1
   // B1 in   RTS in / counter in / free port JP9 (default in)      1
   // B2 out  LED gn (0 = on)                                       0
   // B3 in   comparator LV (Lastspannung)                          0
   // B4 out  LED rd (1 = on)                                       1
   // B7,6,5  SPI1-Master: B7clk, B5mosi must be set as output      101
   // Port B out 2 4 (LEDs) 5 7 SPI
   // PortB B0 is enter in; output 1 for pullup
   //  B2 /OE DI-Leds out0  B4 is redStatusLed out1
   // Port B: B7,6,5 SPI1-Master: B7clk, B5mosi must be set as output
   DDRB = (1<<4) | (1<<2) | (1<<5) | (1<<7); // B4, B2, B5, B7 out
   PORTB = 0x35; // enter, RTS PulllUp, gn out, rd 0n MISO PullUp

   // weAut  Port C
   // C2, C3, C6: /CS=Clk : out 1;
   // C5        : DO enab : out 0;           0: DO disable (LED red)
   // C7   : ether /CS    : out 1;           ether /reset (out 0) on weAut_00
   // C4        : DO OK   : in  1 pullup
   // C01  : Testpin 12   : 1 pullup
   DDRC =  (1<<2) | (1<<3) | (1<<5) | (1<<6) | (1<<7);
   PORTC = ~(_BV(5) );

   // weAut  Port D
   // D7   NIC interupt  : in pullup
   // D6   CTS generator  : out (default)
   // D5   free port on weAut_01 respectively SMC insertion switch
   // D4 3 2  SPI 2 MOSI MISO Clk : 4&3 out
   // D1 0   UART 0 TX RX : 1 out
   DDRD = (1<<1) | (1<<3) | (1<<4) | (1<<6); // D5 not out any more
   PORTD = 0xA6;  // D 7 5 1 2 (16.08.2012 pullup f. mem card)

# else // -- none / most done in bootloader

# endif // correct bootloader only

#elif defined(arduinoUno)

# if  USE_BOOTLOADER < 2
   DDRB  = 0x20;  // PB5 LED yellow (1 on) all other input
   PORTB = 0x20;  // LED on
   DDRC  = 0x00;  // Port C all in as default
   PORTC = 0x0F; // lower nibble pull-up (experimental
   DDRD = 2; // Port D: RxD0, TxD0 1 (out)
   PORTD = 3; // RxD0, TxD0 1 resp pull up.
# else // -- none / most done in bootloader
# endif // correct bootloader only

#elif defined(arduMega2560)

# if  USE_BOOTLOADER < 2
   DDRA = 0;   // basically all input

   DDRB  = 0x80;  // PB7 LED yellow (1 on) all other input
   PORTB = 0x80;  // LED on
   DDRC  = 0x00;  // Port C all in as default
   DDRD = 0; // basically input
   DDRE = 2; // PE0 RxD0,  PE1 TxD0 (out), all other input
   PORTE = 3; // RxD0, TxD0 1 resp. pull-up
   DDRD = 0; DDRE = 0; DDRF = 0;
   DDRG = 0; DDRH = 0; DDRJ = 0;
   DDRK = 0; DDRL = 0;
# else // -- none / most done in bootloader
   DDRC  = 0x00;  // Port C all in as default (bootloader might have set out)
   PORTB = 0x8F;  // LED on and lower 8 pullup

# endif // correct bootloader only

#endif // platform(s)

#if defined(DO_0_PORT)  // lo eight bit process output (DO) by port
   dirPort(DO_0_PORT) = 0xFF;
#endif
#if defined(DI_0_PORT) // lo eight bit process input (Di) by port
   dirPort(DI_0_PORT) = 0x00;
#endif
#if defined(DI_LEDs0_PORT) // lo eight bit process input (Di) Leds by port
   dirPort(DI_LEDs0_PORT) = 0xFF;
#endif
} //   initPorts(void)

void initStatusLeds(uint8_t LEDgnStart, uint8_t LEDrdStart){
#ifdef GN_LED_PORT
   setStatusLedGn(LEDgnStart);
   outPins(GN_LED_PORT, (1<<GN_LED_PIN));

#endif
#ifdef RD_LED_PORT
   setStatusLedRd(LEDrdStart);
   outPins(RD_LED_PORT, (1<<RD_LED_PIN));
#endif
} // initStatusLeds()


/*  \param TP0start initial value for Testpin 0 (ON or OFF)
*  \param TP1start initial value for Testpin 1 (ON or OFF)
*/
void initTestPins(uint8_t TP0start, uint8_t TP1start){
#ifdef TEST_PIN0_PORT
   setTestPin0(TP0start); // reg and pullup
   outPins(TEST_PIN0_PORT, (1<<TEST_PIN0_PIN));
#endif
#ifdef TEST_PIN1_PORT
   setTestPin1(TP1start);
   outPins(TEST_PIN1_PORT, (1<<TEST_PIN1_PIN));
#endif
} // initTestPins()

#if defined(FREE_TIMERS) && FREE_TIMERS > 0
   static struct timer_t freeTims[FREE_TIMERS ];
#endif

/* low level clock / timer / tick initialisation */
void initSystTiming(void){
   OCR0A  = PRESC_TRIM_FAC - 1;
      //   77;   // 78 * 256 = 19968 (best fit to 20000 with 2**n presc.)
      //   61;   // 62 * 256 = 15872 (best <= fit for 16MHz)
   TCCR0A = 0x02; // no output;     WGmode=2 (Bits01)
   TCCR0B = 0x04; // prescaler = 256; WGmode=2 (Bit2)
   OCR0B  = 0xFF; // compare B not used
   TIMSK0 = 0x02; // compare A interrupt enable
   // Vector 17 : Prog.ad. $0020 TIMER0_COMPA Timer0 Compare Match A
   // at 2560 TIMER0_COMPA_vect   is  _VECTOR(21)

   msTimers.next = NULL; // ms timer list
   msTimers.end.v32 = 0;  // ms ticks since start time
   combinedOffset = zoneOffsetSec = 3600;  // + 1 h for CET as default
   isDST = 0;
   secDatTimers.next = NULL;

   secDurTimers.next = NULL;
   secDurTimers.end.v32 =  0; //  seconds since last reset / re-start
   startTime308UTC = DEFAULT_START_TIME;  // secDurTimers.period.v32

   secDatTimers.end.v32 = DEFAULT_START_TIME + 3600; // local time
   msSystClockCount = msAbsClockCount = 0;
} //   initSystTiming(void)


// ------------------ tick interrupt ------------------------------------

uint8_t msIntTick;
uint8_t cn12u8;

uint8_t vcoCount  = VCO_DEFAULT; //!< value controlled oscillator (NTP)
uint8_t vcoSetVal = VCO_DEFAULT; //!< holds best fit for 20 MHz Quartz

/* Trimming of timing oscillator
 *
 *  This function will set the "value controlled oscillator" for the
 *  system's 1 ms tick to normal (usually best) speed.
 */
void setVCOnormal(void){  vcoSetVal = VCO_DEFAULT; }
uint8_t getVCOsetting(void){ return  vcoSetVal; }


/* Trimming of timing oscillator
 *
 *  This function will speed up the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below -200 ms.
 */
void speedVCOup(void){ if (vcoSetVal) ++vcoSetVal; }

/*  Trimming of timing oscillator
 *
 *  This function will slow down the "value controlled oscillator" for the
 *  system's 1 ms tick a little bit. This is for catching up quite little
 *  differences to an externally given absolute (NTP e.g.) time.
 *
 *  Little differences in that sense are below +100 ms.
 */
void slowVCOdown(void){ if (! --vcoSetVal) vcoSetVal = 1; }

/** Tick timer interrupt handler
 *
 *  TIMER0 is \ref intro_sec "weAutSys"' "playground" only. User software
 *  must never ever interfere with it.
 */
ISR(TIMER0_COMPA_vect){
   ++msIntTick; // ++ might not be atomic, but reading 1 byte is
#if USE_TESTPINS == 1
   setTestPin0(msIntTick & 1); // TEST
#endif
   cn12u8 += OCR0A + 1; // advance 16s (16MHz) / 12.5s (20MHz) counter base
   if (! --vcoCount) {  // corriger 0,16 % too fast to 1,12% too slow
#if USE_TESTPINS == 1
      setTestPin1(ON); // TEST
#endif
      vcoCount = vcoSetVal;
      OCR0A  = PRESC_TRIM_FAC;     // count one more
   }  else {   // let it be:  0,018 % too fast
#if USE_TESTPINS == 1
      setTestPin1(OFF); // Test
#endif
      OCR0A  = PRESC_TRIM_FAC - 1; // count normally
   }
   //    SAMPLE_MIN_SP(0);
} // timer 0 match interrupt


/*  Get the 12.8 (16) s tick value (8 bit) */
uint8_t cnt12u8_8(void){ return cn12u8 + TCNT0; }


/*  Go to a label (as value) and stay within the 64Kword page
 *
 *  This procedure is a helper to implement Adam Dumkels' Protothreads with
 *  labels as values on large ATmegas with more than 128K flash.
 *
 *  Note: This can't be used as long as the trampolin approach is used. So
 *  long one has to rely on that linker magic.
 *
 *  \param labelAsValue the target (jump) address' lower 16 bit
 */
void goto_P(void * labelAsValue){
   // if (!labelAsValue) return;
   asm volatile(""
         "IN   r30, __SP_L__  ;  Z = \n\t"
         "IN   r31, __SP_H__  ;  = stackframe \n\t"

#if __AVR_2_BYTE_PC__
         "STD Z+2, %A0     ;  lsb ret add = \n\t"
         "STD Z+1, %B0     ;  msb = label \n\t"
#endif

#if __AVR_3_BYTE_PC__
         "STD Z+3, %A0     ;  lsb ret add = \n\t"
         "STD Z+2, %B0     ;  msb  = label \n\t"
#endif
         :
         : "r" (labelAsValue)
         : "r30", "r31"
   );
} //  goto_P(void *)

/*  Get the upper (bit 16 ...) part of the current address
 *
 *  This function is a helper to implement Adam Dunkels' Protothreads with
 *  labels as values on large ATmegas with more than 128K flash.
 *
 *  \return byte 3 of return address or 0
 */
uint8_t addr3Here(void){
   uint8_t retVal;
   asm volatile(""
         "IN   r30, __SP_L__  ;  Z = \n\t"
         "IN   r31, __SP_H__  ;  = stackframe \n\t"

#if __AVR_2_BYTE_PC__
         "EOR %0, %0       ;  return 0  \n\t"
#endif

#if __AVR_3_BYTE_PC__
         "LDD %0, Z+1     ;  return EIND part \n\t"
#endif
         : "=r" (retVal)
         :
         : "r30", "r31"
   );
   return retVal;
} //  addr3Here()

/*  Get the current address
 *
 *  This function returns the address from where it was called.
 *
 *  \return calling address; the number of relevant bits depends on the flash
 *          size
 */
uint32_t addrHere(void){
   uint32_t retVal;
   asm volatile(""
         "IN   r30, __SP_L__  ;  Z = \n\t"
         "IN   r31, __SP_H__  ;  = stackframe \n\t"
         "MOV  %D0, __zero_reg__;  byte3 = 0 \n\t"
#if __AVR_2_BYTE_PC__
         "MOV  %C0, __zero_reg__;  byte2 = 0 \n\t"
         "LDD %B0, Z+1     ;  ret add msb \n\t"
         "LDD %A0, Z+2     ;  ret add lsb \n\t"
#endif

#if __AVR_3_BYTE_PC__
         "LDD %C0, Z+1     ;  ret add EIND \n\t"
         "LDD %B0, Z+2     ;  ret add msb \n\t"
         "LDD %A0, Z+3     ;  ret add lsb \n\t"
#endif
         : "=r" (retVal)
         :
         : "r30", "r31"
   );


   return retVal;
}
// ------------------   basic control functions   --------------------------

// internal: moved here from ntp.c due to the damned
// inline asm bug 48435

#include "we-aut_sys/ntp.h"

/*  Difference of two NTP time stamps
 *
 *  This function calculates *result = * a - *b; <br />
 *  No pointer must be null, but two of them might point to the
 *  same \ref ntpTimestamp_t structure.
 *
 *  \param a points to first operand (not NULL)
 *  \param b   points to second operand (not NULL)
 *  \param result points to result structure (not NULL)
 */
void ntpTimestampDif(struct ntpTimestamp_t * result,
                        struct ntpTimestamp_t * a,
                           struct ntpTimestamp_t * b){
   asm volatile(""
         "adiw r26, 7 ; x[7]    (2C) \n\t"

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP == 0
         "ldd r24, %a1+7  ;  a*  (2C) \n\t"
         "ldd r25, %a2+7  ;  b*  (2C) \n\t"
         "sub r24, r25    ;  a* - b*  \n\t"
         "st  X, r24      ;  result* = (2C)  \n\t"
#else
         "st  X,  __zero_reg__  ; NTP_FRACT_DROP  1..3  \n\t"
         "clc                   ; clear carry \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 2
         "ldd r24, %a1+6  ;  a* next byte  (2C)  \n\t"
         "ldd r25, %a2+6  ;  b* next byte   (2C) \n\t"
         "sbc r24, r25    ;  a - b - carry   \n\t"
         "st  -X, r24     ; result next byte (2C) \n\t"
#else
         "st  -X,  __zero_reg__  ; NTP_FRACT_DROP  2..3  \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 3
         "ldd r24, %a1+5   \n\t"
         "ldd r25, %a2+5  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"
#else
         "st  -X,  __zero_reg__  ; NTP_FRACT_DROP  3  \n\t"
#endif

         "ldd r24, %a1+4   \n\t"
         "ldd r25, %a2+4  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ldd r24, %a1+3   \n\t"
         "ldd r25, %a2+3  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ldd r24, %a1+2   \n\t"
         "ldd r25, %a2+2  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ldd r24, %a1+1   \n\t"
         "ldd r25, %a2+1  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ld  r24, %a1   \n\t"
         "ld  r25, %a2  \n\t"
         "sbc r24, r25   \n\t"
         "st  -X, r24   \n\t"
        :
        : "x"(result), "y"(a), "z"(b)
        : "r24", "r25"
       );
}  // ntpTimestampDif(...)



