/**
 *  \file modbus.c
 *
 *  weAutSys Modbus server implementation
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include "we-aut_sys/common.h"
#include "we-aut_sys/modbus.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/log_streams.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/utils.h"

#include "uip/uip.h"

#include <string.h>

#if DEBUG_MOD
//                            0123456789.
INFLASH(char const  modbMark[])   = " ## Modbus ";
INFLASH(char const  modbRequ[])   = " ## ModRequ";
INFLASH(char const  modbResp[])   = " ## ModResp";
INFLASH(char const  modbErrR[])   = " ## ModErrR";
//INFLASH(char const  modbDO[])     = " # TelnDO ";
//INFLASH(char const  modbWILL[])   = " # TelnWI ";
//INFLASH(char const  modbDONT[])   = " # TelnDN ";
//INFLASH(char const  modbWONT[])   = " # TelnDN ";
//INFLASH(char const  modbSB[])     = " # TelnSB ";
#endif

/* System Modbus configuration */
struct modConfData_t systModConf = {.port = 0};

/*  Initialise the Modbus server
 *
 *  This function initialises the Modbus server. It is (indirectly) called
 *  whenever the Ethernet link goes up and hence usually after reset.
 *  This implementation does nothing.
 */
void modbusInit(void){
   if (!systModConf.port)  {
      memset(&systModConf, 0, sizeof(struct modConfData_t)); // appModFun, options, unit
      systModConf.port = MODBUS_PORT;
   }
} // modbusInit()

/*  Register the application Modbus handler function
 *
 * \param appModFun pointer to the function to register; NULL to de-register
 * \sa modConfData_t
 */
void registerAppModFun(p2ptFunM appModFun){
   systModConf.appModFun = appModFun;
} // registerAppModFun(p2ptFunM)


/*  Handle Modbus  events
 *
 *  This function handles Modbus events. It is a as a protothread (part of)
 *  the Modbus server state machine.
 *
 *  Hint: This implements at at present the basic communication and a CLI
 *  (command line interpreter).
 *
 *  The current appstate.flag usage is: <br />
 *  Bit 0 (1): command interpreter active <br />
 *  Bit 1 (2): prompt output (after last output) pending <br />
 *  Bit 3 (4): consume and send the pending \ref bufLogStreams output
 *           (before prompt). <br />
 */
ptfnct_t modbusAppcall(void){

   struct thr_data_t * const s = (struct thr_data_t *)&(uip_conn->appstate);
   struct  modThr_data_t * const m = &(s->modThrData);

   if (uip_connected()) { // newly connected  ----------------------------
#if DEBUG_MOD
      //           0123456789.123456789v123456 789t
      char tS[] = " ## Modbus conn .... .... \n";
      fourHexs(&tS[16], (uint16_t)bufLogCons);
      fourHexsBE(&tS[21], uip_conn->rport);
      bufLogTxt(tS, 27);
#endif
    // todo actions on new Modbus connect (if any)
      return PT_EXITED;
   } // ---------------     newly connected   -------------------- end ----

   if (uip_closed()) { // closed by other host (client)   -----------------
      // todo actions on a Modbus disconnect (if any)
      // uip_close();  // not necessary
#if DEBUG_MOD
         bufLogHex_P(modbMark, 0xFC); // foreign client has closed
         bufLogPut2C(' ','\n');
#endif
         PT_LEAVE(&s->pt);
   }  // --------------   closed by other host (client)  ---------- end ----

   if (uip_aborted()) { // aborted by other host (client) ------------------
      // todo actions on a Modbus disconnect (if any)
      uip_close();  // not necessary
#if DEBUG_MOD
         bufLogHex_P(modbMark, 0xFA); // foreign client has aborted
         bufLogPut2C(' ','\n');
#endif
         PT_LEAVE(&s->pt);
   }  // -------------     aborted by other host (client) --------- end ----


   if (uip_newdata() ) {  //= = = = = = = = = = =  =  new  data  begin  ====
      const uint16_t gotLen = uip_datalen();
      if (gotLen < 10) { // MBAB 7 + min 3 for smallest PDU possible
         return PT_EXITED; // we just ignore too short telegrams so far
      }

      // set up modThrData
      memset(m, 0, sizeof(struct modThr_data_t));
      memcpy(&(m->config), &systModConf, sizeof(struct modConfData_t));
      m->retC = PT_EXITED;
      m->gotLen = gotLen;
      const uint8_t telStartLen = gotLen > sizeof(struct modTelegr_t)
                            ? sizeof(struct modTelegr_t) : (uint8_t)gotLen;
      memcpy(&(m->reqModTelegr), uip_appdata, telStartLen);
      const uint8_t fcFlag = m->fcFlag = ((uint8_t *)uip_appdata)[MODB_FCIND];


#if DEBUG_MOD
      bufLogTxt_P(modbRequ);
      char tmp[58];
      formModTelegr(tmp, &(m->reqModTelegr));
      bufLogTxt(tmp, 58);
      bufLogPutC('\n');
      //           0123456789.123456789v123 4 567 8
      //char tS[] = " ## ModbusRq .... fx.. \n";
      //fourHexsBE(&tS[13], *((uint16_t *)uip_appdata)); // trans ID
      //twoHexs(&tS[20], fcFlag);  // funCode
      //bufLogTxt(tS, 24);
#endif
      // check and prepare the function codes
      switch (fcFlag) {
         ucnt16_t tmp;
         case WRITE_COIL: // Write single coil (0x05)
            m->dataModel = COIL_SINGLE; // coils
            tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
            tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
            m->inDatTarg.v16 = tmp.v16; // target address = Bit number
            m->inData = (uint8_t *)uip_appdata + 10; // 00: clear; FF: set
            break;

         case WRITE_COILS: //  (0x0F)
            m->dataModel = COILS_BYMAPD; // coils

            tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
            if (tmp.v8[0] & 7) { // bit address within byte
               m->errorCode = MODB_EXC_ADDR; // invalid address
               goto weDoSendErrResp;
            }
            tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
            m->inDatTarg.v16 = tmp.v16 / 8;

            tmp.v8[0] = m->reqModTelegr.pduCont[1].v8[1];
            if (tmp.v8[0] & 7) { // bit count modulo 8 not 0
               m->errorCode = MODB_EXC_ADDR; // invalid address
                goto weDoSendErrResp;
            }
            tmp.v8[1] = m->reqModTelegr.pduCont[1].v8[0];
            m->inDataByteCnt =  (uint8_t)(tmp.v16 / 8);
            m->inData = (uint8_t *)uip_appdata + 12;
            break;

         case READ_COILS: // (0x01)
            m->dataModel = COILS_BYMAPD; // coils
            goto setReadBitValues; // >=====|
            break;

         case READ_DISCRETE_INPUTS: // (0x02)
            m->dataModel = DISC_INP_BYMAPD;
            setReadBitValues:      // <<==== |
            tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
            if (tmp.v8[0] & 7) { // bit address within byte
               m->errorCode = MODB_EXC_ADDR; // invalid address
                goto weDoSendErrResp;
            }
            tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
            m->outDatSou.v16 = tmp.v16 / 8;

            tmp.v8[0] = m->reqModTelegr.pduCont[1].v8[1];
            if (tmp.v8[0] & 7) { // bit count modulo 8 not 0
               m->errorCode = MODB_EXC_ADDR; // invalid address
               goto weDoSendErrResp;
            }
            tmp.v8[1] = m->reqModTelegr.pduCont[1].v8[0];
            m->outDataByteCnt = (uint8_t)(tmp.v16 / 8);
            m->outData = (uint8_t *)uip_appdata + 9;
            break;

            // end of bitwise sh.t  (greetings from KOP FUB AWL) ----------

         case WRITE_HOLDING_REGISTER: // (0x06)
            m->dataModel = HLD_REGISTERS;
            m->inData = uip_appdata + MODB_FCIND + 3;
            m->inDataByteCnt = 2;
            goto setWriteHldValues; // >=====|
            break;

         case WRITE_HOLDING_REGISTERS: // (0x10)
            m->dataModel = HLD_REGISTERS;
            m->inData = uip_appdata + MODB_FCIND + 6;
            m->inDataByteCnt = (m->reqModTelegr.pduCont[1].v8[1]) * 2;
            setWriteHldValues:      // <<==== |
            tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
            tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
            m->inDatTarg.v16 = tmp.v16 * 2; // target BYTE address
            if (m->reqModTelegr.pduCont[1].v8[0]) {  // count too large
               m->errorCode = MODB_EXC_DATA; // data error
               goto weDoSendErrResp;
            }  // count too large
#if DEBUG_MOD
   //           0123456789.123456789v123456 789t12 3
   char tS[] = " ## ModbWrHldR m- w - t - \n";
   tS[16] = m->dataModel;
   twoHexs(&tS[23], m->inDatTarg.v8[0]);
   twoHexs(&tS[19], m->inDataByteCnt);
   bufLogTxt(tS, 27);
#endif
            break;

         case READ_INPUT_REGISTERS:  // (0x04)
            m->dataModel = INP_REGISTERS;
            goto setReadWordValues; // >=====|
            break;

         case READ_HOLDING_REGISTERS: // (0x03)
            m->dataModel = HLD_REGISTERS;
            setReadWordValues:     // <<==== |
            tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
            tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
            m->outDatSou.v16 = tmp.v16 * 2; // source BYTE address
            if (m->reqModTelegr.pduCont[1].v8[0]) {  // count too large
               m->errorCode = MODB_EXC_DATA; // data error
               goto weDoSendErrResp;
            }  // count too large
            m->outDataByteCnt = (m->reqModTelegr.pduCont[1].v8[1]) * 2;
            m->outData = uip_appdata + MODB_FCIND + 2;
#if DEBUG_MOD
    //           0123456789.123456789v123456 789t12 3
   char tS2[] = " ## ModbRdHldR m- s - r - \n";
   tS2[16] = m->dataModel;
   twoHexs(&tS2[19], m->outDatSou.v8[0]);
   twoHexs(&tS2[23], m->outDataByteCnt);
   bufLogTxt(tS2, 27);
#endif
            break;


         case WRITE_READ_REGISTERS: // (0x17, 23)
            m->dataModel = HLD_REGISTERS;
            // write
            if (m->reqModTelegr.pduCont[3].v8[0]) {  // count too large
                m->errorCode = MODB_EXC_DATA; // data error
                goto weDoSendErrResp;
             }  // count too large
            m->inDataByteCnt = (m->reqModTelegr.pduCont[3].v8[1]) * 2;
            tmp.v8[0] = m->reqModTelegr.pduCont[2].v8[1];
            tmp.v8[1] = m->reqModTelegr.pduCont[2].v8[0];
            m->inDatTarg.v16 = tmp.v16 * 2; // target BYTE address

            // read

             tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
             tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
             m->outDatSou.v16 = tmp.v16 * 2; // source BYTE address
             if (m->reqModTelegr.pduCont[1].v8[0]) {  // count too large
                m->errorCode = MODB_EXC_DATA; // data error
                goto weDoSendErrResp;
             }  // count too large
             m->outDataByteCnt = (m->reqModTelegr.pduCont[1].v8[1]) * 2;
             m->outData = uip_appdata + MODB_FCIND + 2;
#if DEBUG_MOD
   //            0123456789.123456789v123456789t12345 6 7
   char tS3[] = " ## ModbWrRdHl m- s - r -  t - w - \n";
   tS3[16] = m->dataModel;
   twoHexs(&tS3[19], m->outDatSou.v8[0]);
   twoHexs(&tS3[23], m->outDataByteCnt);

   twoHexs(&tS3[28], m->inDatTarg.v8[0]);
   twoHexs(&tS3[32], m->inDataByteCnt);
   bufLogTxt(tS3, 36);
#endif
            break;

         case MASK_WRITE_REGISTER: // (0x16, 22)
             m->dataModel = HLD_MASK;
             m->inData = uip_appdata + MODB_FCIND + 3; // address of AND
             m->inDataByteCnt = 2;
             tmp.v8[0] = m->reqModTelegr.pduCont[0].v8[1];
             tmp.v8[1] = m->reqModTelegr.pduCont[0].v8[0];
             m->inDatTarg.v16 = tmp.v16 * 2; // target BYTE address
             if (m->reqModTelegr.pduCont[1].v8[0]) {  // count too large
                m->errorCode = MODB_EXC_DATA; // data error
                goto weDoSendErrResp;
             }  // count too large
 #if DEBUG_MOD
    //            0123456789.123456789v12 3456 789t12 3
    char tS4[] = " ## ModbMskHld m- t - \n";
    tS4[16] = m->dataModel;
    twoHexs(&tS4[19], m->inDatTarg.v8[0]);
    bufLogTxt(tS4, 23);
 #endif
             break;

      } // switch (fcFlag)


      weCalltheUserHandler: ;
      const p2ptFunM appModFunPt = m->config.appModFun;
      if (!appModFunPt) {
         m->errorCode = 1; // invalid / unimplemented function
      } else {
         (*appModFunPt)(m); // call the registered application handler
      }

    //  weDoTheEnd:                // <<=== come from done directly good
      if (m->errorCode)  { // end with error
         m->fcFlag = 0; // function performed
         weDoSendErrResp:          // <<=== come from done directly good
         m->reqModTelegr.fCode |= 0x80;
         m->reqModTelegr.pduCont[0].v8[0] = m->errorCode;
         m->reqModTelegr.byToFol.v16 = HTONS(3);
#if DEBUG_MOD
         bufLogTxt_P(modbErrR);
         char tmp[58];
         formModTelegr(tmp, &(m->reqModTelegr));
         bufLogTxt(tmp, 58);
         bufLogPutC('\n');
#endif
         uip_send(&(m->reqModTelegr), 9); // answer with error
         PT_EXIT(&s->pt);
      } // end with error

      if (m->retC < PT_EXITED) PT_EXIT(&s->pt); // leave w/o send (block,yield)
      ucnt16_t sendLen, copyLen;
      copyLen.v16 = sendLen.v16 = m->gotLen; // prepare for echo

      switch (m->fcFlag) { // prepare OK response
         case READ_COILS:             // fx01
         case READ_DISCRETE_INPUTS:   // fx02
         case READ_HOLDING_REGISTERS: // fx03
         case READ_INPUT_REGISTERS:   // fx04
            copyLen.v16 = MODB_MBAB_LEN + 1; // just until fc
            const uint8_t t230 =  m->outDataByteCnt;
            * (((uint8_t *) uip_appdata) + MODB_FCIND + 1) = t230;
            sendLen.v16 =  MODB_MBAB_LEN + 2 + t230;
            break;
         // end input (they look alike, as we map bit mod 8 byte)

         case WRITE_COIL:              // fx05
         case WRITE_HOLDING_REGISTER:  // fx06
         case WRITE_COILS:             // fx0F
         case WRITE_HOLDING_REGISTERS: // fx10
            copyLen.v16 = sendLen.v16 = MODB_MBAB_LEN + 5; // just repeat start of req
            break;

         case MASK_WRITE_REGISTER:     // fx16
            copyLen.v16 = sendLen.v16 = MODB_MBAB_LEN + 7; // just repeat start of req
            break;

         case WRITE_READ_REGISTERS:    // fx17
            copyLen.v16 = MODB_MBAB_LEN + 9; // just until fc
            const uint8_t t251 = m->outDataByteCnt;
            * (((uint8_t *) uip_appdata) + MODB_FCIND + 9) = t251;
            sendLen.v16 =  MODB_MBAB_LEN + 10 + t251;
            break;

      } // switch (m->fcFlag) (prepare OK response)

      ucnt16_t t277;
      t277.v16 =  sendLen.v16 - 6; // bytes to follow in correct byte order
      m->reqModTelegr.byToFol.v8[0] = t277.v8[1]; // Modbus sh..
      m->reqModTelegr.byToFol.v8[1] = t277.v8[0]; // big endian
      memcpy(uip_appdata, &(m->reqModTelegr), copyLen.v16); // copy request back

#if DEBUG_MOD >= 3
         bufLogTxt_P(modbResp);
         //  char tmp[58];
         formModTelegr(tmp, &(m->reqModTelegr));
         bufLogTxt(tmp, 58);
         bufLogPutC('\n');
#endif

      uip_send(uip_appdata, sendLen.v16); // just echo (TEST only)
      // todo actions on new data

      PT_EXIT(&s->pt);
   } //= = = = = = = = = = = = = = = = = = = = =  =  new  data  (end)  ====


   if (uip_poll() ) { //= = = = = = = = = = = = = = =   poll     begin ====
      const uint8_t fcFlag = m->fcFlag;
      const uint8_t prevRetC = m->retC;
      if (fcFlag || prevRetC < PT_EXITED) return PT_EXITED;
#if DEBUG_MOD
      const uint8_t pollCnt = ++  m->pollCnt;
#endif
      /// PT_BEGIN(&s->pt);  // standard schedule (re-) entry point of thread
      // todo actions on poll

#if DEBUG_MOD
      //           0123456789.123456789v123 4 567 8
      char tS[] = " ## ModbusPoll .. fx.. \n";
      twoHexs(&tS[15], pollCnt);
      twoHexs(&tS[20], fcFlag);
      bufLogTxt(tS, 24);
#endif
      goto weCalltheUserHandler;

      /// PT_EXIT(&s->pt);  // standard schedule (re-) entry point of thread
   } //= = = = = = = = = = = = = = = = = = = = =  =  poll     (end)  ====

   // neither connected nor newData, poll or closed
   // what else ? uip_aborted()
   return PT_EXITED;
} // modbusAppcall()

