/**
 * \file ntp.c
 *
 * weAutSys NTP (client) implementation
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include "we-aut_sys/common.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/ntp.h"
#include "we-aut_sys/enc28j60.h"

#include "we-aut_sys/utils.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/log_streams.h"

#include "uip/uip.h"
#include <string.h>

struct ntpState_t ntpState; // the NTP state

/** Set a NTP time stamp to actual time
 *
 *  \param ntpTimestamp the stamp to be set (never null)
 */
void setNTPstampAct(struct ntpTimestamp_t * ntpTimestamp){
   ucnt32_t st;
   st.v32 = secTime308UTC()  + OFFSET2NTPTIME;
   ntpTimestamp->seconds.v8[3] =  st.v8[0];
   ntpTimestamp->seconds.v8[2] =  st.v8[1];
   ntpTimestamp->seconds.v8[1] =  st.v8[2];
   ntpTimestamp->seconds.v8[0] =  st.v8[3];

   ntpTimestamp->fraction.v16[0] = msAbsClockCount; // 16 bit 0..999 in litlE

   asm volatile("" // adjust ms to fract and litlE to BigE
         "ldd r24, Z+4 ; lower ms 8 bits   \n\t"
         "ldd r25, Z+5 ; higher ms 2 bits   \n\t"
         "movw r22, r24 ; save for + 1/64  \n\t"

         "lsr r25  ;   bit 0 out \n\t"
         "ror r24  ;   bit 0 out bit 0 in \n\t"
         "ror r25  ;   bit 1 out bit 0 in \n\t"
         "ror r24  ;   bit 1 out bit 1 in \n\t"
         "ror r25  ;   bit 1 in \n\t"

         "add r25, r22 ; + 1/64 lo \n\t"
         "adc r24, r23 ; hi  (1/1024)*(1 + 1/64) \n\t"

         "std Z+4, r24   ;   \n\t"
         "std Z+5, r25   ;   \n\t"
         "std Z+6,  __zero_reg__   ;   \n\t"
         "std Z+7,  __zero_reg__   ;   \n\t"

         :
         : "z"(ntpTimestamp)
         : "r24", "r25", "r22", "r23"
   );
} // setNTPstampAct()


//--------------------------   algorithm implementation   ------------------

/*  Half a NTP time stamp value
 *
 *  This function calculates *result = *result / 2 <br />
 *
 *  \param result points to operand and result structure (not NULL)
 */
void ntpTimestampHalf(struct ntpTimestamp_t * result){
  asm volatile(""
        "ld r24, Z  ; lo is MSB (big endian)   \n\t"
        "asr r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"

        "ld r24, Z  ; next 1  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"

        "ld r24, Z  ; next 2  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"

        "ld r24, Z  ; next 3  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"

        "ld r24, Z  ; next 4  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 3
        "ld r24, Z  ; next 5  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"
#else
         "st  Z+,  __zero_reg__  ; NTP_FRACT_DROP  1..3  \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 2
        "ld r24, Z  ; next 6  \n\t"
        "ror r24    ;  \n\t"
        "st Z+,r24  ;   \n\t"
#else
         "st  Z+,  __zero_reg__  ; NTP_FRACT_DROP  1..3  \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP == 0
        "ld r24, Z  ; last 7  \n\t"
        "ror r24    ;  \n\t"
        "st Z,r24   ;   \n\t"
#else
         "st  Z,  __zero_reg__  ; NTP_FRACT_DROP  1..3  \n\t"
#endif
       :
       : "z"(result)
       : "r24"
      );
}  // ntpTimestampHalf(...)


//--------------------------   protocol implementation   ------------------


INFLASH(char const ntpInitMes[])   = "\n  NTP init'd., state: ";
INFLASH(char const ntpResetMes[])  = "\n  NTP reset'd.  \n";
//                              0 123456789.1234567 8 9 .
INFLASH(char const ntpRequestMes[])= "\n  NTP request  ";
INFLASH(char const ntpReceiveMes[])= "\n  NTP receive  ";
INFLASH(char const ntpDeltaFrMes[])= "\n  NTP delta_f  ";
INFLASH(char const ntpServRcvMes[])=   "  NTP srv rec  ";
INFLASH(char const ntpAppCallMes[])= "\n  NTP appCall  ";
//  INFLASH(char const ntpClockAdMes[])= "\n  NTP ClckAdj. ";


/*  Adjust the actual (system) time by a NTP time stamp difference
 *
 *  This function adjusts the internal date and time (\ref secTime308UTC())
 *  by an amount given as a a NTP time stamp \ref ntpTimestampDif "difference"
 *  to the\ref secTime308Loc "actual time in s", \ref msAbsClockCount "ms".
 *
 *  Big (>=2**31 in the seconds part) differences will quite naturally act as
 *  negative in the sense of \ref secTime308Loc "second based" timing's
 *  (and NTP's) modulo  2**32 arithmetic.
 *
 *  This function should be used for any absolute date time adjustments caused
 *  by NTP client functions (or may be also other time sources).
 *
 *  The action taken and the flags set in .lastAdjust depend on the difference
 *  amount<ul>
 *  <li> 0 ms : no action; </li>
 *  <li> -100 .. +50ms : no setting;
 *       only oscillator speed up / slow down </li>
 *  <li> -999 .. + 999ms : setting of time,
 *       oscillator speed up / slow down </li>
 *  <li> above -1 .. +1s :  setting of time </li>
 *
 *  \param ntpTimeDif  the adjustment difference in seconds,fraction; big
 *                values act as negative in the sense of seconds
 *                modulo 2**32 arithmetic
 */
void adjustNTPtime308UTC(struct ntpTimestamp_t * ntpTimeDif){

   ntpState.adjuState = (ntpState.adjuState >> 4)  // last >> previous
         | NTP_ADJUSTAT_RNG_HS;                     // preset seconds high
   ntpState.msDifPrev = ntpState.msDif;
   ntpTimestampToMillies(&ntpState.msDif, ntpTimeDif);

   uint8_t secIsM1 = 0; // .seconds == -1 is assumed false

   if (ntpTimeDif->seconds.v8[0] & 0x80 ) { // negative
      for(;;) { // check negative ranges
         if (ntpTimeDif->seconds.v8[0] != 0xFF) break;
         if (ntpTimeDif->seconds.v8[1] != 0xFF) break;
         if (ntpTimeDif->seconds.v8[2] != 0xFF) break;
         if (ntpTimeDif->seconds.v8[3] != 0xFF) break;
         secIsM1 = 0x33;
         ntpState.adjuState &= NTP_ADJUSTAT_PRV_MSK; // clear outside second

         if (ntpTimeDif->fraction.v8[0] != 0xFF) {
            if (ntpTimeDif->fraction.v8[0] < 0xE6) {
               ntpState.adjuState |= NTP_ADJUSTAT_RNG_HM; // set hi ms range
            }
            break;
         }
         return; // consider below 8 ms as jitter
      } // check negative ranges
      ntpState.adjuState |= NTP_ADJUSTAT_MINUS; // set negative flag
      if(ntpState.adjuState & NTP_ADJUSTAT_PRV_PLUS) { // change of sign
         setVCOnormal();
      } else if (! (ntpState.adjuState & NTP_ADJUSTAT_RNG_1S)) { // inside second
         if (ntpState.msDif < ntpState.msDifPrev) slowVCOdown(); // slow down
      }
   } else { // 0 or positive
      for(;;) {  // check positive ranges
         if (ntpTimeDif->seconds.v8[0]) break;
         if (ntpTimeDif->seconds.v8[1]) break;
         if (ntpTimeDif->seconds.v8[2]) break;
         if (ntpTimeDif->seconds.v8[3]) {
            if (ntpTimeDif->seconds.v8[3] == 1) {
               ntpState.adjuState = (ntpState.adjuState
                       & NTP_ADJUSTAT_PRV_MSK) // clear ..
                           |  NTP_ADJUSTAT_RNG_1S; // set 1..2 s range
            }
            break;
         }
         ntpState.adjuState &= NTP_ADJUSTAT_PRV_MSK; // clear outside second
         if (ntpTimeDif->fraction.v8[0]) {
            if (ntpTimeDif->fraction.v8[0] > 0x0C) {
               ntpState.adjuState |= NTP_ADJUSTAT_RNG_HM; // set hi ms range
            }
            break;
         }
         return; // consider below 8 ms as jitter
      } // check positive ranges
      ntpState.adjuState |= NTP_ADJUSTAT_PLUS; // set positive flag
      if(ntpState.adjuState & NTP_ADJUSTAT_PRV_MINUS) { // change of sign
         setVCOnormal();
      } else if (! (ntpState.adjuState & NTP_ADJUSTAT_RNG_1S)) { // inside second
         if (ntpState.msDif > ntpState.msDifPrev) speedVCOup(); // speed up
      }
   } // // 0 or positive

   if (!(ntpState.adjuState & NTP_ADJUSTAT_RNGMSK)) return; // lo ms -> no action

   setVCOnormal(); // set to normal speed after hard set (experimental)

   if (adjustUTCcount
         && (ntpState.state  & NTP_STATE_OPMSK) == NTP_STATE_CONTSY )  {
      if (!ntpState.count++) return; // ignore first big dif as spike
   }

   msAbsClockCount += ntpState.msDif;
   if (msAbsClockCount >= 1000) {
      msAbsClockCount -= 1000;
      if (secIsM1) { // was -1 now inside second
         ntpState.adjuState &= ~ NTP_ADJUSTAT_RNG_1S ; // inside 1 second
         if (! (++adjustUTCcount)) adjustUTCcount = 1;
         return; // no further action
      } //  // was -1 now inside second
      secIsM1 = 1; // one to add
   } else {
      if (!(ntpState.adjuState & NTP_ADJUSTAT_RNG_1S)) return; // inside 1s
      secIsM1 = 0; // signal nothing to add
   }

   // inline asm the following start ??
   ntpState.secDif = convert32endian(ntpTimeDif->seconds.v32);
   if (secIsM1) ++ntpState.secDif;
   secDatTimers.end.v32 += ntpState.secDif; // local time in seconds since 1st ...
   if(! adjustUTCcount) {
      startTime308UTC  += ntpState.secDif; // secDurTimers.period.v32
   }
   if (ntpState.secDif == 1) {
      sys1sPeriodFlag |= ABSsecADJ; // flag 1s system thread to run (date, timers)
   } else {
      sys1sPeriodFlag |= ABSsecADJ | ABSsecSET;
   }

   if (ntpState.logLev >= optVerboseNum) { // info on NTP
      bufLogDecHex_P(ntpDeltaFrMes+1, ntpState.msDif, ntpState.secDif);
   } // // info on NTP

   if (! (++adjustUTCcount)) adjustUTCcount = 1;
} // adjustNTPtime308UTC(struct ntpTimestamp_t *)



/*  This helper function makes the first basic message in uip_appdata:
 *  a NTP client request with almost all fields 0 and then has it sent.
 *  Set to 0 are: stratum (don't know), poll, rootDelay, rootDispersion
 *  referenceIdentifier, referenceTimestamp.seconds, originateTimestamp
 *  receiveTimestamp
 */
static void sendNTPrequest(void){
  struct ntpMess_t *m = (struct  ntpMess_t *)uip_appdata;
  memset(m, 0, sizeof(struct ntpMess_t) - 8); // -8: transmit TStamp is last

  /*  replaced by memset -----
  m-.fraction.v32 = 0;    ---- replaced by memset()   */

  m->flags = LI_ALARM | NTP_VERSION  | NTP_CLIENT;
  m->precision = -6;     // 2**-6 see setNTPstampAct()
  setNTPstampAct(&ntpState.lastSendTimestamp);  // we stamp now (not the ENC)
  m->transmitTimestamp = ntpState.lastSendTimestamp;
  uip_send(uip_appdata, sizeof(struct ntpMess_t));
  if (ntpState.logLev >= optVerboseNum) { // info on NTP
      bufLogDecHex_P(ntpRequestMes, msAbsClockCount,  // trAnsmit
                 convert32endian(ntpState.lastSendTimestamp.fraction.v32));

     /// bufLogTxt_P(ntpRequestMes);
     /// bufLog4HexBE(ntpState.lastSendTimestamp.fraction.v16[0]);
     /// bufLogPut2C(' ', '\n'); // 1 blank 1 LF
  } // info on NTP
} //  sendNTPrequest(void)


/*  Reset the NTP (client) to initial state
 *
 *  This function initialises the NTP structures and (protothread) state
 *  machine.
 *  This function resets all NTP state to initial. In contrast to
 *  \ref ntpInit no trial to connect or bind to NTP ports or servers. The
 *  protocol is not (not even indirectly) started.
 */
void ntpReset(void){
   ntpState.state = NTP_STATE_RESET;
   ntpState.adjuState = 0;
   ntpState.count  = 0;
   if (ntpState.conn) { // had connection
      ntpState.conn->lport = 0; // release local port
      ntpState.conn  = NULL;
   }
   ntpState.msDif  = 0;
   PT_INIT(&ntpState.pt);
   if (ntpState.logLev >= optQuietNum) { // info on NTP
      bufLogTxt_P(ntpResetMes); // reset
   }
} // ntpReset()

#define ALLOW_ENCRES2NTP 1 //!< Allow NIC reset to NTP state machine \showinitializer

/* Initialise the NTP (client)
 *
 * This function initialises the NTP structures and state machine
 * (= protothread). The protocol is not started by this function; the
 * work is done by later calls of \ref ntpAppcall().
 *
 * \param pref2 if true use second NTP server (if available)
 */
void ntpInit(uint8_t pref2){
   // ntpReset(); // is (to be) done by system or application before
   if (!(curIpConf.useFlags & NTP_CLIENT_MSK)) return; // be not client

   initTimer(&ntpState.timer, 1, sec_dur_TIMER_TYPE);

   if (!(curIpConf.setFlags & NTP_MSK)) return; // no server
   if (pref2) {
      if (curIpConf.setFlags & NTP2_MSK) {
         ntpState.conn = uip_udp_new(&curIpConf.ntp2Addr, HTONS(NTP_PORT));
      } else {
         pref2 = 0; // no server 2
      }
   }
   if (! pref2) {
      ntpState.conn = uip_udp_new(&curIpConf.ntpAddr, HTONS(NTP_PORT));
   }

   if(ntpState.conn != NULL) {
      uip_udp_bind(ntpState.conn, HTONS(NTP_PORT));
      ntpState.state = pref2 ? NTP_STATE_CONN2 : NTP_STATE_CONN1;
   }  else {
      ntpState.state = NTP_STATE_CNPROB;
   }
#if DEBUG_NTP && DEBUG_NTP <= 4 && DEBUG_NTP >= 1
   if (!ntpState.logLev) ntpState.logLev = optSilentNum + DEBUG_NTP;
#endif
   if (ntpState.logLev > optSilentNum) { // info on NTP
      bufLogHex_P(ntpInitMes, ntpState.state); // connect to ..
      bufLogPut2C(' ','\n');
      bufLogDate();
   } // info on NTP
} // ntpInit()


/** Check the NTP server's answer */
uint8_t parseNTPreply(void){
   struct ntpMess_t *m = (struct  ntpMess_t *)uip_appdata;
   if (uip_datalen() != sizeof(struct ntpMess_t)) {
      if (ntpState.logLev > optVerboseNum) { // info on NTP
         bufLogHex_P(ntpReceiveMes, 0xB0); // bogus wrong length
      }
      return  0;   // bogus wrong length
   }
   if (memcmp(&ntpState.lastSendTimestamp, &m->originateTimestamp, 8)) { // T1
      if (ntpState.logLev > optVerboseNum) { // info on NTP
         bufLogHex_P(ntpReceiveMes, 0xB1); // bogus wrong time
      }
      return  0;   // bogus wrong time
   }
   if ((m->flags &  FLAGS_LI_MASK) == LI_ALARM) { // server has no time
      if (ntpState.logLev > optVerboseNum) { // info on NTP
         bufLogHex_P(ntpReceiveMes, 0xBC); // bogus LI alarm no server time
      }
      return  0;   // server has no time
   }


   // we (have to) stamp now. The ENC can't
   setNTPstampAct(&ntpState.lastReceiveTimestamp); // stamp T4 now
   if (ntpState.logLev > optVerboseNum) { // info on NTP
      bufLogDec_P(ntpServRcvMes, msAbsClockCount);
      bufLog8HexBE(ntpState.lastReceiveTimestamp.seconds.v32);
      bufLog8HexBE(ntpState.lastReceiveTimestamp.fraction.v32);
      bufLogLF(1);
   //   bufLogDecHex_P(ntpReceiveMes+1, 14, msAbsClockCount,  // trAnsmit
     //        convert32endian(ntpState.lastReceiveTimestamp.fraction.v32));
   }

   ntpTimestampDif(&ntpState.lastT21, &m->receiveTimestamp,
                    &ntpState.lastSendTimestamp);  // T21 = T2 - T1
   ntpTimestampDif(&ntpState.lastT34, &m->transmitTimestamp,
                    &ntpState.lastReceiveTimestamp);  // T34 = T3 - T4

   if (ntpState.logLev > optVerboseNum) { // info on NTP
      uint16_t fracMS;
      ntpTimestampToMillies(&fracMS, &m->transmitTimestamp);
         bufLogDecHex_P(ntpReceiveMes +1, fracMS,
                         convert32endian(m->transmitTimestamp.fraction.v32));
   }

   return 1;
} // parseNTPreply()


/* Handle NTP (server) events
 *
 * This function handles NTP events. And it is a (as a protothread) the
 * NTP client state machine.
 */
ptfnct_t ntpAppcall(void){
   if (ntpState.logLev > optInfoNum) { // info on NTP
      if (uip_flags & ~(UIP_POLL | UIP_NEWDATA)) { // neither poll nor new
         bufLogTxt_P(ntpAppCallMes);
         char tmp[3];
         twoHexs(tmp, uip_flags);
         bufLogTxt(tmp, 2);
         bufLogPut2C(' ', '\n'); // 1 blank 1 LF
      } // neither poll nor new: very strange
   }

   // do nothing on reset state machine
   const uint8_t curSrv = ntpState.state & NTP_STATE_SRVMSK; // server bits
   if (!curSrv)  PT_EXIT(&ntpState.pt);  // no connect yet
   // to here only after successful   ntpInit()

   PT_BEGIN(&ntpState.pt);

   ntpState.state = curSrv | NTP_STATE_TRYREQ;  // set is first request

   for(;;) { // endless
      // wait just for test
      restartTimer(&ntpState.timer);  // start time out is ~1s
      PT_WAIT_UNTIL(&ntpState.pt, timerLapsed(ntpState.timer));

      sendNTPrequest();
      ntpState.state |= NTP_STATE_W4REPL; // wait for reply (please priority)
      restartTimer(&ntpState.timer);  // start time out is ~1s
      PT_WAIT_UNTIL(&ntpState.pt, uip_newdata()
                                             || timerLapsed(ntpState.timer));
      if (!uip_newdata()) { // no reception (time out)
         ntpState.state = curSrv | NTP_STATE_CNPROB;  // probl. connection
         if(!(ntpState.timer.period.v8[0] & 0xC0)) {
            ntpState.timer.period.v8[0] += 3; // increase  to 34s
         }
         if (++ntpState.count > 7) { // end of patience
#if ALLOW_ENCRES2NTP  // NTP state machine may reset the NIC (i.e. ENC28J60)
            ntpReset(); // sets timer back to 1s
            encReset();
            PT_WAIT_UNTIL(&ntpState.pt, timerLapsed(ntpState.timer));
            encReset();
            PT_WAIT_UNTIL(&ntpState.pt, timerLapsed(ntpState.timer)); // twice
            encInit();
            PT_WAIT_UNTIL(&ntpState.pt, timerLapsed(ntpState.timer));
#endif // NTP may reset the NIC as ENC28J60 probably stopped receiving
            ntpInit(curSrv != NTP_STATE_CONN2); // resets this PT
            break; // out for retry (endless) with both servers
         } // end of patience
         continue;  // retry with current set server
      } // no reception (time out)

      if (!parseNTPreply()) {
         ntpState.state = curSrv | NTP_STATE_SEPROB;
         ntpState.timer.period.v8[0] = 9; // every nine sec
         if (++ntpState.count > 7) { // end of patience
            ntpReset();
            ntpInit(curSrv != NTP_STATE_CONN2); // resets this PT
            return PT_EXITED; // but retry endless with both servers
         } // end of patience
         continue; // just retry endless so far
      }
      ntpState.state &= ~NTP_STATE_W4REPL; // no more waiting wait for reply
      if ((ntpState.state &  NTP_STATE_OPMSK) != NTP_STATE_CONTSY) {
         ntpState.state = curSrv | NTP_STATE_CNETAB;
         ntpState.count = 0;
      } // first establish

            // calculate delta
      struct ntpTimestamp_t  delta;
      ntpTimestampSum(&delta, &ntpState.lastT34, &ntpState.lastT21);
      ntpTimestampHalf(&delta);

      adjustNTPtime308UTC(&delta);
      if (!(ntpState.adjuState &  NTP_ADJUSTAT_RNG_1S)) { // adjust < 1 s
         ntpState.state = curSrv | NTP_STATE_CONTSY;
         ntpState.count = 0;
         ntpState.timer.period.v8[0] = 31; // every.. sec
      } else {
         ntpState.timer.period.v8[0] = 13; // every.. sec
      }

   } // endless ---  while ntpState.state <= NTP_STATE_CONN2

   PT_END(&ntpState.pt)
} // ntpAppcall()

